/*
	by Vladimir Smagin, 2020
	vlad@blindage.org
*/

package hetzner

import (
	"context"
	"fmt"
	"os"
	"strings"

	hclouddns "git.blindage.org/21h/hcloud-dns"
	log "github.com/sirupsen/logrus"
	"sigs.k8s.io/external-dns/endpoint"
	"sigs.k8s.io/external-dns/plan"
	"sigs.k8s.io/external-dns/provider"
)

const (
	hetznerCreate = "CREATE"
	hetznerDelete = "DELETE"
	hetznerUpdate = "UPDATE"
	hetznerTTL    = 600
)

type HetznerChanges struct {
	Action            string
	ZoneID            string
	ZoneName          string
	ResourceRecordSet hclouddns.HCloudRecord
}

type HetznerProvider struct {
	Client       hclouddns.HCloudDNS
	domainFilter endpoint.DomainFilter
	DryRun       bool
}

func NewHetznerProvider(ctx context.Context, domainFilter endpoint.DomainFilter, dryRun bool) (*HetznerProvider, error) {
	token, ok := os.LookupEnv("HETZNER_TOKEN")
	if !ok {
		return nil, fmt.Errorf("No token found")
	}

	provider := &HetznerProvider{
		Client:       *hclouddns.New(token),
		domainFilter: domainFilter,
		DryRun:       dryRun,
	}
	return provider, nil
}

func (p *HetznerProvider) Records(ctx context.Context) ([]*endpoint.Endpoint, error) {
	zones, err := p.Client.GetZones(hclouddns.HCloudGetZonesParams{})
	if err != nil {
		return nil, err
	}
	endpoints := []*endpoint.Endpoint{}
	for _, zone := range zones.Zones {
		records, err := p.Client.GetRecords(hclouddns.HCloudGetRecordsParams{ZoneID: zone.ID})
		if err != nil {
			return nil, err
		}

		for _, r := range records.Records {
			if provider.SupportedRecordType(string(r.RecordType)) {
				name := r.Name + "." + zone.Name

				if r.Name == "@" {
					name = zone.Name
				}

				endpoints = append(endpoints, endpoint.NewEndpoint(name, string(r.RecordType), r.Value))
			}
		}
	}

	return endpoints, nil
}

func (p *HetznerProvider) ApplyChanges(ctx context.Context, changes *plan.Changes) error {
	combinedChanges := make([]*HetznerChanges, 0, len(changes.Create)+len(changes.UpdateNew)+len(changes.Delete))

	combinedChanges = append(combinedChanges, p.newHetznerChanges(hetznerCreate, changes.Create)...)
	combinedChanges = append(combinedChanges, p.newHetznerChanges(hetznerUpdate, changes.UpdateNew)...)
	combinedChanges = append(combinedChanges, p.newHetznerChanges(hetznerDelete, changes.Delete)...)

	return p.submitChanges(ctx, combinedChanges)
}

func (p *HetznerProvider) submitChanges(ctx context.Context, changes []*HetznerChanges) error {
	if len(changes) == 0 {
		log.Infof("All records are already up to date")
		return nil
	}

	zones, err := p.Client.GetZones(hclouddns.HCloudGetZonesParams{})
	if err != nil {
		return err
	}

	zoneChanges := p.seperateChangesByZone(zones.Zones, changes)

	for _, changes := range zoneChanges {
		for _, change := range changes {

			log.WithFields(log.Fields{
				"record":  change.ResourceRecordSet.Name,
				"type":    change.ResourceRecordSet.RecordType,
				"ttl":     change.ResourceRecordSet.TTL,
				"action":  change.Action,
				"zone":    change.ZoneName,
				"zone_id": change.ZoneID,
			}).Info("Changing record.")

			change.ResourceRecordSet.Name = strings.TrimSuffix(change.ResourceRecordSet.Name, "."+change.ZoneName)
			if change.ResourceRecordSet.Name == change.ZoneName {
				change.ResourceRecordSet.Name = "@"
			}
			if change.ResourceRecordSet.RecordType == endpoint.RecordTypeCNAME {
				change.ResourceRecordSet.Value += "."
			}

			switch change.Action {
			case hetznerCreate:
				record := hclouddns.HCloudRecord{
					RecordType: change.ResourceRecordSet.RecordType,
					ZoneID:     change.ZoneID,
					Name:       change.ResourceRecordSet.Name,
					Value:      change.ResourceRecordSet.Value,
					TTL:        change.ResourceRecordSet.TTL,
				}
				_, err := p.Client.CreateRecord(record)
				if err != nil {
					return err
				}
			case hetznerDelete:
				_, err := p.Client.DeleteRecord(change.ResourceRecordSet.ID)
				if err != nil {
					return err
				}
			case hetznerUpdate:
				record := hclouddns.HCloudRecord{
					RecordType: change.ResourceRecordSet.RecordType,
					ZoneID:     change.ZoneID,
					Name:       change.ResourceRecordSet.Name,
					Value:      change.ResourceRecordSet.Value,
					TTL:        change.ResourceRecordSet.TTL,
				}
				_, err := p.Client.UpdateRecord(record)
				if err != nil {
					return err
				}
			}
		}
	}

	return nil
}

func (p *HetznerProvider) newHetznerChanges(action string, endpoints []*endpoint.Endpoint) []*HetznerChanges {
	changes := make([]*HetznerChanges, 0, len(endpoints))
	ttl := hetznerTTL
	for _, e := range endpoints {

		if e.RecordTTL.IsConfigured() {
			ttl = int(e.RecordTTL)
		}

		change := &HetznerChanges{
			Action: action,
			ResourceRecordSet: hclouddns.HCloudRecord{
				RecordType: hclouddns.RecordType(e.RecordType),
				Name:       e.DNSName,
				Value:      e.Targets[0],
				TTL:        ttl,
			},
		}
		changes = append(changes, change)
	}
	return changes
}

func (p *HetznerProvider) seperateChangesByZone(zones []hclouddns.HCloudZone, changes []*HetznerChanges) map[string][]*HetznerChanges {
	change := make(map[string][]*HetznerChanges)
	zoneNameID := provider.ZoneIDName{}

	for _, z := range zones {
		zoneNameID.Add(z.ID, z.Name)
		change[z.ID] = []*HetznerChanges{}
	}

	for _, c := range changes {
		zoneID, zoneName := zoneNameID.FindZone(c.ResourceRecordSet.Name)
		if zoneName == "" {
			log.Debugf("Skipping record %s because no hosted zone matching record DNS Name was detected", c.ResourceRecordSet.Name)
			continue
		}
		c.ZoneName = zoneName
		c.ZoneID = zoneID
		change[zoneID] = append(change[zoneID], c)

	}
	return change
}
